#!/usr/bin/env python
# coding=utf-8
import os, pickle
import numpy as np
from . import utils


def prepare_tfrecord(raw_data_file,
                     model,
                     token2id_dct,
                     tokenize,
                     preprocess_raw_data_fn,
                     save_data_prefix,
                     save_data_dir='../data',
                     update_txt=False,
                     update_tfrecord=False,
                     **kwargs):
    # 1. read raw_data_file
    # 2. preprocess (seg/split...) -> .train .dev .test
    # 3. generate tfrecord
    # 4. load tfrecord
    if not os.path.exists(save_data_dir):
        os.makedirs(save_data_dir)
    # define data file name
    train_txt_file = f'{save_data_dir}/{save_data_prefix}_train_data.txt'
    dev_txt_file = f'{save_data_dir}/{save_data_prefix}_dev_data.txt'
    test_txt_file = f'{save_data_dir}/{save_data_prefix}_test_data.txt'
    train_tfrecord_file = f'{save_data_dir}/{save_data_prefix}_train_data.tfrecord'
    dev_tfrecord_file = f'{save_data_dir}/{save_data_prefix}_dev_data.tfrecord'
    test_tfrecord_file = f'{save_data_dir}/{save_data_prefix}_test_data.tfrecord'

    if update_txt:
        update_tfrecord = True
        if os.path.exists(train_txt_file): os.remove(train_txt_file)
        if os.path.exists(dev_txt_file): os.remove(dev_txt_file)
        if os.path.exists(test_txt_file): os.remove(test_txt_file)

    if not all([os.path.exists(train_tfrecord_file), os.path.exists(dev_tfrecord_file)]) or update_tfrecord:  # 没有或需要更新tfrecord
        # 首先检查txt file
        if not all([os.path.exists(train_txt_file), os.path.exists(dev_txt_file)]):
            train_data, dev_data, test_data = preprocess_raw_data_fn(raw_data_file, tokenize=tokenize, token2id_dct=token2id_dct, **kwargs)
            if train_data:
                print(f'generating train txt file... filename: {train_txt_file}')
                utils.list2file(train_txt_file, train_data)
            if dev_data:
                print(f'generating dev txt file... filename: {dev_txt_file}')
                utils.list2file(dev_txt_file, dev_data)
            if test_data:
                print(f'generating test txt file... filename: {test_txt_file}')
                utils.list2file(test_txt_file, test_data)

        if os.path.exists(train_txt_file):
            model.generate_tfrecord(train_txt_file, token2id_dct, train_tfrecord_file)
        if os.path.exists(dev_txt_file):
            model.generate_tfrecord(dev_txt_file, token2id_dct, dev_tfrecord_file)
        if os.path.exists(test_txt_file):
            model.generate_tfrecord(test_txt_file, token2id_dct, test_tfrecord_file)

    return train_tfrecord_file, dev_tfrecord_file, test_tfrecord_file


def prepare_pkldata(raw_data_file,
                    model,
                    token2id_dct,
                    tokenize,
                    preprocess_raw_data_fn,
                    save_data_prefix,
                    save_data_dir='../data',
                    update_txt=False,
                    update_pkl=False,
                    **kwargs):
    # 返回data: 各个字段有全量数据
    # 同时返回训练及验证的数据数量
    if not os.path.exists(save_data_dir):
        os.makedirs(save_data_dir)
    # define data file name
    train_txt_file = f'{save_data_dir}/{save_data_prefix}_train_data.txt'
    dev_txt_file = f'{save_data_dir}/{save_data_prefix}_dev_data.txt'
    test_txt_file = f'{save_data_dir}/{save_data_prefix}_test_data.txt'
    train_pkl_file = f'{save_data_dir}/{save_data_prefix}_train_data.pkl'
    dev_pkl_file = f'{save_data_dir}/{save_data_prefix}_dev_data.pkl'
    test_pkl_file = f'{save_data_dir}/{save_data_prefix}_test_data.pkl'

    if update_txt:
        update_pkl = True
        if os.path.exists(train_txt_file): os.remove(train_txt_file)
        if os.path.exists(dev_txt_file): os.remove(dev_txt_file)
        if os.path.exists(test_txt_file): os.remove(test_txt_file)

    if not all([os.path.exists(train_pkl_file), os.path.exists(dev_pkl_file)]) or update_pkl:  # 没有或需要更新pkldata
        # 首先检查txt file
        if not all([os.path.exists(train_txt_file), os.path.exists(dev_txt_file)]):
            train_data, dev_data, test_data = preprocess_raw_data_fn(raw_data_file, tokenize=tokenize, token2id_dct=token2id_dct, **kwargs)
            if train_data:
                utils.list2file(train_txt_file, train_data)
                print(f'generate train txt file ok! {train_txt_file}')
            if dev_data:
                utils.list2file(dev_txt_file, dev_data)
                print(f'generate dev txt file ok! {dev_txt_file}')
            if test_data:
                utils.list2file(test_txt_file, test_data)
                print(f'generate test txt file ok! {test_txt_file}')

        if os.path.exists(train_txt_file):
            train_pkldata = model.generate_data(train_txt_file, token2id_dct)
            pickle.dump(train_pkldata, open(train_pkl_file, 'wb'))
            print(f'generate and save train pkl file ok! {train_pkl_file}')
        if os.path.exists(dev_txt_file):
            dev_pkldata = model.generate_data(dev_txt_file, token2id_dct)
            pickle.dump(dev_pkldata, open(dev_pkl_file, 'wb'))
            print(f'generate and save train pkl file ok! {dev_pkl_file}')
        if os.path.exists(test_txt_file):
            test_pkldata = model.generate_data(test_txt_file, token2id_dct)
            pickle.dump(test_pkldata, open(test_pkl_file, 'wb'))
            print(f'generate and save train pkl file ok! {test_pkl_file}')
            
    return train_pkl_file, dev_pkl_file, test_pkl_file


# 得到当前图中所有变量的名称
# tensor_names = [tensor.name for tensor in graph.as_graph_def().node]  # 得到当前图中所有变量的名称
# for tensor_name in tensor_names:
#     if not tensor_name.startswith('save') and not tensor_name.startswith('gradients'):
#         if 'Adam' not in tensor_name and 'Initializer' not in tensor_name:
#             print(tensor_name)


def gen_pos_neg_sample(items, sample_idx, num_neg_exm=9, seed=1234):
    import random
    import copy
    total_exm = []
    random.seed(seed)  # 采样随机负样本的随机性
    cands_ids = list(range(len(items)))

    for i, exm in enumerate(items):
        exm.append(1)  # add pos label
        total_exm.append(exm)  # add pos exm

        neg_exm_lst = []
        neg_sample_set = set()

        while len(neg_exm_lst) < min(num_neg_exm, len(items) - 1):  # data数量比负采样样本数还少的情况
            neg_id = random.choice(cands_ids)
            if neg_id == i:  # 如果采到自己
                continue
            if items[neg_id][sample_idx] == exm[sample_idx]:  # 如果sample的文本相同
                continue
            if items[neg_id][sample_idx] in neg_sample_set:  # 如果sample已经采样了
                continue
            neg_sample_set.add(items[neg_id][sample_idx])
            neg_exm = copy.deepcopy(exm)
            neg_exm[sample_idx] = items[neg_id][sample_idx]  # change to neg 
            neg_exm[-1] = 0  # add neg label
            neg_exm_lst.append(neg_exm)  # add neg exm
          
        total_exm.extend(neg_exm_lst)  # add neg exms
        
    return total_exm


def calc_recall(epo_s1, epo_prob, epo_y, topk=5, strip_pad=False):
    comb = list(zip(epo_s1, epo_prob, epo_y))
    s1_dct = {}
    for s1, prob, y in comb:
        if 'ndarray' in str(type(s1)):
            s1 = s1.flatten().tolist()
        if strip_pad:
            while s1[-1] == 0:
                s1.pop()
        s1 = tuple(s1)
        if s1 not in s1_dct:
            s1_dct[s1] = []
        s1_dct[s1].append([prob, y])
    for s1 in s1_dct:
        s1_dct[s1].sort(key=lambda e: e[0], reverse=True)
    total_recall = []
    for s1 in s1_dct:
        y_lst = [e[1] for e in s1_dct[s1]]
        # print(len(y_lst))
        if len(y_lst) < 2:  # 没有正负样本
            continue
        # print(sum(y_lst))
        if sum(y_lst) != 1:  # 真实标签中没有正样本1 或超过1个正样本
            continue
        recall_item = [sum(y_lst[:topk]) for topk in range(1, topk+1)]
        total_recall.append(recall_item)
    total_recall = np.mean(total_recall, axis=0).tolist()
    return total_recall


import math
import numpy as np
from scipy.stats import multivariate_normal
import tensorflow as tf

def learning_rate(step_num, d_model=512, warmup_steps=4000):
    a = step_num**(-0.5)
    b = step_num*warmup_steps**(-1.5)
    return a, b, d_model**(-0.5) * min(step_num**(-0.5), step_num*(warmup_steps**(-1.5))) 

def selu(x):
    alpha = 1.6732632423543772848170429916717
    scale = 1.0507009873554804934193349852946
    print('use selu')
    return scale*tf.where(x>=0.0, x, alpha*tf.nn.elu(x))

def bilinear_sim_4d(x, y, is_nor=True):
    '''calulate bilinear similarity with two 4d tensor.
    
    Args:
        x: a tensor with shape [batch, time_x, dimension_x, num_stacks]
        y: a tensor with shape [batch, time_y, dimension_y, num_stacks]

    Returns:
        a tensor with shape [batch, time_x, time_y, num_stacks]

    Raises:
        ValueError: if
            the shapes of x and y are not match;
            bilinear matrix reuse error.
    '''
    M = tf.get_variable(
        name="bilinear_matrix", 
        shape=[x.shape[2], y.shape[2], x.shape[3]],
        dtype=tf.float32,
        initializer=tf.orthogonal_initializer())
    sim = tf.einsum('biks,kls,bjls->bijs', x, M, y)

    if is_nor:
        scale = tf.sqrt(tf.cast(x.shape[2] * y.shape[2], tf.float32))
        scale = tf.maximum(1.0, scale)
        return sim / scale
    else:
        return sim


def bilinear_sim(x, y, is_nor=True):
    '''calculate bilinear similarity with two tensor.
    Args:
        x: a tensor with shape [batch, time_x, dimension_x]
        y: a tensor with shape [batch, time_y, dimension_y]
    
    Returns:
        a tensor with shape [batch, time_x, time_y]
    Raises:
        ValueError: if
            the shapes of x and y are not match;
            bilinear matrix reuse error.
    '''
    M = tf.get_variable(
        name="bilinear_matrix", 
        shape=[x.shape[-1], y.shape[-1]],
        dtype=tf.float32,
        initializer=tf.orthogonal_initializer())
    sim = tf.einsum('bik,kl,bjl->bij', x, M, y)

    if is_nor:
        scale = tf.sqrt(tf.cast(x.shape[-1] * y.shape[-1], tf.float32))
        scale = tf.maximum(1.0, scale)
        return sim / scale
    else:
        return sim

def dot_sim(x, y, is_nor=True):
    '''calculate dot similarity with two tensor.

    Args:
        x: a tensor with shape [batch, time_x, dimension]
        y: a tensor with shape [batch, time_y, dimension]
    
    Returns:
        a tensor with shape [batch, time_x, time_y]
    Raises:
        AssertionError: if
            the shapes of x and y are not match.
    '''
    assert x.shape[-1] == y.shape[-1]

    sim = tf.einsum('bik,bjk->bij', x, y)

    if is_nor:
        scale = tf.sqrt(tf.cast(x.shape[-1], tf.float32))
        scale = tf.maximum(1.0, scale)
        return sim / scale
    else:
        return sim

def layer_norm(x, axis=None, epsilon=1e-6):
    '''Add layer normalization.

    Args:
        x: a tensor
        axis: the dimensions to normalize

    Returns:
        a tensor the same shape as x.

    Raises:
    '''
    print('wrong version of layer_norm')
    scale = tf.get_variable(
        name='scale',
        shape=[1],
        dtype=tf.float32,
        initializer=tf.ones_initializer())
    bias = tf.get_variable(
        name='bias',
        shape=[1],
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    if axis is None:
        axis = [-1]

    mean = tf.reduce_mean(x, axis=axis, keep_dims=True)
    variance = tf.reduce_mean(tf.square(x - mean), axis=axis, keep_dims=True)
    norm = (x-mean) * tf.rsqrt(variance + epsilon)
    return scale * norm + bias

def layer_norm_debug(x, axis = None, epsilon=1e-6):
    '''Add layer normalization.

    Args:
        x: a tensor
        axis: the dimensions to normalize

    Returns:
        a tensor the same shape as x.

    Raises:
    '''
    if axis is None:
        axis = [-1]
    shape = [x.shape[i] for i in axis]

    scale = tf.get_variable(
        name='scale',
        shape=shape,
        dtype=tf.float32,
        initializer=tf.ones_initializer())
    bias = tf.get_variable(
        name='bias',
        shape=shape,
        dtype=tf.float32,
        initializer=tf.zeros_initializer())

    mean = tf.reduce_mean(x, axis=axis, keep_dims=True)
    variance = tf.reduce_mean(tf.square(x - mean), axis=axis, keep_dims=True)
    norm = (x-mean) * tf.rsqrt(variance + epsilon)
    return scale * norm + bias

def dense(x, out_dimension=None, add_bias=True):
    '''Add dense connected layer, Wx + b.

    Args:
        x: a tensor with shape [batch, time, dimension]
        out_dimension: a number which is the output dimension

    Return:
        a tensor with shape [batch, time, out_dimension]

    Raises:
    '''
    if out_dimension is None:
        out_dimension = x.shape[-1]

    W = tf.get_variable(
        name='weights',
        shape=[x.shape[-1], out_dimension],
        dtype=tf.float32,
        # initializer=tf.orthogonal_initializer())
        initializer=tf.contrib.layers.xavier_initializer())
    if add_bias:
        bias = tf.get_variable(
            name='bias',
            shape=[1],
            dtype=tf.float32,
            initializer=tf.zeros_initializer())
        return tf.einsum('bik,kj->bij', x, W) + bias
    else:
        return tf.einsum('bik,kj->bij', x, W)

def matmul_2d(x, out_dimension, drop_prob=None):
    '''Multiplies 2-d tensor by weights.

    Args:
        x: a tensor with shape [batch, dimension]
        out_dimension: a number

    Returns:
        a tensor with shape [batch, out_dimension]

    Raises:
    '''
    W = tf.get_variable(
        name='weights',
        shape=[x.shape[1], out_dimension],
        dtype=tf.float32,
        initializer=tf.orthogonal_initializer())
    if drop_prob is not None:
        W = tf.nn.dropout(W, drop_prob)
        print('W is dropout')

    return tf.matmul(x, W)

def gauss_positional_encoding_vector(x, role=0, value=0):
    position = int(x.shape[1])
    dimension = int(x.shape[2])
    print('position: %s' %position)
    print('dimension: %s' %dimension)

    _lambda = tf.get_variable(
        name='lambda',
        shape=[position],
        dtype=tf.float32,
        initializer=tf.constant_initializer(value))
    _lambda = tf.expand_dims(_lambda, axis=-1)

    mean = [position/2.0, dimension/2.0]

    #cov = [[position/3.0, 0], [0, dimension/3.0]]
    sigma_x = position/math.sqrt(4.0*dimension)
    sigma_y = math.sqrt(dimension/4.0)
    cov = [[sigma_x*sigma_x, role*sigma_x*sigma_y], 
            [role*sigma_x*sigma_y, sigma_y*sigma_y]]

    pos = np.dstack(np.mgrid[0:position, 0:dimension])

    
    rv = multivariate_normal(mean, cov)
    signal = rv.pdf(pos) 
    signal = signal - np.max(signal)/2.0

    signal = tf.multiply(_lambda, signal)
    signal = tf.expand_dims(signal, axis=0)

    print('gauss positional encoding')

    return x + _lambda * signal

def positional_encoding(x, min_timescale=1.0, max_timescale=1.0e4, value=0):
    '''Adds a bunch of sinusoids of different frequencies to a tensor.

    Args:
        x: a tensor with shape [batch, length, channels]
        min_timescale: a float
        max_timescale: a float

    Returns:
        a tensor the same shape as x.

    Raises:
    '''
    length = x.shape[1]
    channels = x.shape[2]
    _lambda = tf.get_variable(
        name='lambda',
        shape=[1],
        dtype=tf.float32,
        initializer=tf.constant_initializer(value))

    position = tf.to_float(tf.range(length))
    num_timescales = channels // 2
    log_timescale_increment = (
        math.log(float(max_timescale) / float(min_timescale)) /
        (tf.to_float(num_timescales) - 1))
    inv_timescales = min_timescale * tf.exp(
        tf.to_float(tf.range(num_timescales)) * -log_timescale_increment)
    scaled_time = tf.expand_dims(position, 1) * tf.expand_dims(inv_timescales, 0)
    signal = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=1)
    signal = tf.pad(signal, [[0, 0], [0, tf.mod(channels, 2)]])
    #signal = tf.reshape(signal, [1, length, channels])
    signal = tf.expand_dims(signal, axis=0)

    return x + _lambda * signal


def positional_encoding_vector(x, min_timescale=1.0, max_timescale=1.0e4, value=0):
    '''Adds a bunch of sinusoids of different frequencies to a tensor.

    Args:
        x: a tensor with shape [batch, length, channels]
        min_timescale: a float
        max_timescale: a float

    Returns:
        a tensor the same shape as x.

    Raises:
    '''
    length = x.shape[1]
    channels = x.shape[2]
    _lambda = tf.get_variable(
        name='lambda',
        shape=[length],
        dtype=tf.float32,
        initializer=tf.constant_initializer(value))
    _lambda = tf.expand_dims(_lambda, axis=-1)

    position = tf.to_float(tf.range(length))
    num_timescales = channels // 2
    log_timescale_increment = (
        math.log(float(max_timescale) / float(min_timescale)) /
        (tf.to_float(num_timescales) - 1))
    inv_timescales = min_timescale * tf.exp(
        tf.to_float(tf.range(num_timescales)) * -log_timescale_increment)
    scaled_time = tf.expand_dims(position, 1) * tf.expand_dims(inv_timescales, 0)
    signal = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=1)
    signal = tf.pad(signal, [[0, 0], [0, tf.mod(channels, 2)]])

    signal = tf.multiply(_lambda, signal)
    signal = tf.expand_dims(signal, axis=0)

    return x + signal

def mask(row_lengths, col_lengths, max_row_length, max_col_length):
    '''Return a mask tensor representing the first N positions of each row and each column.

    Args:
        row_lengths: a tensor with shape [batch]
        col_lengths: a tensor with shape [batch]

    Returns:
        a mask tensor with shape [batch, max_row_length, max_col_length]

    Raises:
    '''
    row_mask = tf.sequence_mask(row_lengths, max_row_length) #bool, [batch, max_row_len]
    col_mask = tf.sequence_mask(col_lengths, max_col_length) #bool, [batch, max_col_len]

    row_mask = tf.cast(tf.expand_dims(row_mask, -1), tf.float32)
    col_mask = tf.cast(tf.expand_dims(col_mask, -1), tf.float32)

    return tf.einsum('bik,bjk->bij', row_mask, col_mask)

def weighted_sum(weight, values):
    '''Calcualte the weighted sum.

    Args:
        weight: a tensor with shape [batch, time, dimension]
        values: a tensor with shape [batch, dimension, values_dimension]

    Return:
        a tensor with shape [batch, time, values_dimension]

    Raises:
    '''
    return tf.einsum('bij,bjk->bik', weight, values)


        




#!/usr/bin/env python
# coding=utf-8
import tensorflow as tf
from tensorflow.python.util import nest
from .beam_search import *
from .embedding import *
from .common_layers import *


def transformer_encoder(encoder_input, encoder_valid_mask,
                        hidden_size=512,
                        filter_size=2048,
                        num_heads=8,
                        num_encoder_layers=6,
                        dropout=0.,
                        attention_dropout=0.,
                        relu_dropout=0.,
                        ):
    # attention mask
    encoder_pad_mask = 1. - encoder_valid_mask  # 1 for pad 0 for nonpad 用于ffn  [batch,len]
    encoder_self_attention_bias = encoder_pad_mask * -1e9  # attention mask
    encoder_self_attention_bias = tf.expand_dims(tf.expand_dims(encoder_self_attention_bias, axis=1), axis=1)  # [batch,1,1,length]

    x = encoder_input
    with tf.variable_scope('encoder'):
        for layer in range(num_encoder_layers):
            with tf.variable_scope('layer_%d' % layer):
                with tf.variable_scope('self_attention'):
                    y = multihead_attention(
                        layer_preprocess(x, dropout_rate=dropout),
                        None,  # self attention
                        encoder_self_attention_bias,
                        hidden_size,
                        hidden_size,
                        hidden_size,
                        num_heads,
                        attention_dropout)
                    x = layer_postprecess(x, y, dropout_rate=dropout)
                with tf.variable_scope('ffn'):
                    y = transformer_ffn_layer(
                        layer_preprocess(x, dropout_rate=dropout),
                        filter_size,
                        hidden_size,
                        relu_dropout=relu_dropout,
                        pad_mask=encoder_pad_mask)
                    x = layer_postprecess(x, y, dropout_rate=dropout)
        encoder_output = layer_preprocess(x, dropout_rate=dropout)
    return encoder_output


def transformer_decoder(decoder_input, encoder_output, decoder_valid_mask, encoder_valid_mask,
                        cache=None,
                        hidden_size=512,
                        filter_size=2048,
                        num_heads=8,
                        num_decoder_layers=6,
                        dropout=0.,
                        attention_dropout=0.,
                        relu_dropout=0.,
                        ):
    # attention mask
    encoder_pad_mask = 1. - encoder_valid_mask  # [batch,length] 1 for pad 0 for nonpad
    encoder_decoder_attention_bias = encoder_pad_mask * -1e9  # attention mask
    encoder_decoder_attention_bias = tf.expand_dims(tf.expand_dims(encoder_decoder_attention_bias, axis=1), axis=1)  # [batch,1,1,length]

    if decoder_valid_mask is None:  # infer mode  decoder_input [batch,1,embed]
        decoder_pad_mask = None
        decoder_self_attention_bias = tf.zeros([1, 1, 1, 1])  # infer阶段其实不需要mask
    else:  # train model
        decoder_pad_mask = 1. - decoder_valid_mask  # [batch,length] 1 for pad; 0 for nonpad 用于ffn
        decoder_length = shape_list(decoder_input)[1]
        decoder_self_attention_bias = attention_bias_lower_triangle(decoder_length)  # 下三角 causal attention

    x = decoder_input
    with tf.variable_scope('decoder'):
        for layer in range(num_decoder_layers):
            layer_name = 'layer_%d' % layer
            layer_cache = cache[layer_name] if cache is not None else None
            with tf.variable_scope(layer_name):
                with tf.variable_scope('self_attention'):
                    y = multihead_attention(
                        layer_preprocess(x, dropout_rate=dropout),
                        None,
                        decoder_self_attention_bias,
                        hidden_size,
                        hidden_size,
                        hidden_size,
                        num_heads,
                        attention_dropout,
                        cache=layer_cache)
                    x = layer_postprecess(x, y, dropout_rate=dropout)
                with tf.variable_scope('encdec_attention'):
                    y = multihead_attention(
                        layer_preprocess(x, dropout_rate=dropout),
                        encoder_output,
                        encoder_decoder_attention_bias,
                        hidden_size,
                        hidden_size,
                        hidden_size,
                        num_heads,
                        attention_dropout,
                        cache=layer_cache)
                    x = layer_postprecess(x, y, dropout_rate=dropout)
                with tf.variable_scope('ffn'):
                    y = transformer_ffn_layer(
                        layer_preprocess(x, dropout_rate=dropout),
                        filter_size,
                        hidden_size,
                        relu_dropout=relu_dropout,
                        pad_mask=decoder_pad_mask)
                    x = layer_postprecess(x, y, dropout_rate=dropout)
        decoder_output = layer_preprocess(x, dropout_rate=dropout)
    return decoder_output


def attention_base_pool(inputs, input_valid_mask,
                        name='sent_pool', reuse=tf.AUTO_REUSE,
                        attn_v=None):
    """ sentence vector """
    # inputs [batch,len,hidden]
    # input_valid_mask [batch,len]
    # attn_v if exist [hidden]
    hidden_size = shape_list(inputs)[-1]
    if attn_v is None:
        with tf.variable_scope(name, reuse=reuse):
            attn_v = tf.get_variable('attention_v', [hidden_size],  # [hidden]
                                     initializer=tf.contrib.layers.xavier_initializer())

    # attention alpha before normalizer
    score = tf.reduce_sum(inputs * attn_v, axis=-1)  # [batch,len]
    input_pad_mask = 1. - input_valid_mask  # 1 for pad 0 for nonpad
    attention_mask = input_pad_mask * -1e9  # attention mask

    # normalizer
    alignments = tf.nn.softmax(score + attention_mask)  # [batch,len]

    outputs = tf.reduce_sum(inputs * tf.expand_dims(alignments, axis=-1), axis=1)  # [batch,hidden]

    return outputs


def multi_head_attention_base_pool(inputs, input_valid_mask,
                                   total_key_size,
                                   total_value_size,
                                   num_heads,
                                   name='sent_pool',
                                   reuse=tf.AUTO_REUSE,
                                   dropout=0.,
                                   attn_v=None):
    """ sentence vector """
    # inputs [batch,len,hidden]
    # return [batch,hidden]
    hidden_size = shape_list(inputs)[-1]
    if attn_v is None:
        with tf.variable_scope(name, reuse=reuse):
            attn_v = tf.get_variable('attention_v', [hidden_size],  # [hidden]
                                     initializer=tf.contrib.layers.xavier_initializer())

    encoder_pad_mask = 1. - input_valid_mask  # 1 for pad 0 for nonpad [batch,len]
    bias = encoder_pad_mask * -1e9  # attention mask
    bias = tf.expand_dims(tf.expand_dims(bias, axis=1), axis=1)  # [batch,1,1,length]

    attn_v = tf.reshape(attn_v, [1, 1, hidden_size])  # [1,1,hidden]

    # inputs -> k,v  attn_v -> q
    q = tf.layers.dense(attn_v, total_key_size, use_bias=False, name='q')  # [1,1,key_size]
    q = tf.tile(q, [tf.shape(inputs)[0], 1, 1])  # [batch,1,key_size]

    k = tf.layers.dense(inputs, total_key_size, use_bias=False, name='k')  # [batch,len,key_size]
    v = tf.layers.dense(inputs, total_value_size, use_bias=False, name='v')  # [batch,len,key_size]

    # split head
    q = split_heads(q, num_heads)  # [batch,head,1,hidden-]s
    k = split_heads(k, num_heads)  # [batch,head,len,hidden-]
    v = split_heads(v, num_heads)  # [batch,head,len,hidden-]

    key_size_per_head = total_key_size // num_heads
    q = q * key_size_per_head ** -0.5  # scale

    # q * k_T * v
    x = dot_product_attention(q, k, v, bias, dropout_rate=dropout)  # [batch,head,1,hidden]

    x = combine_heads(x)  # [batch,1,hidden]

    x.set_shape(x.shape.as_list()[:-1] + [total_key_size])  # set last dim specifically

    x = tf.squeeze(x, 1)

    x = tf.layers.dense(x, total_value_size, use_bias=False, name='output_transform')  # [batch,hiden]

    return x


def mean_pool(inputs, input_valid_mask, name='mean_pool'):
    # inputs [batch,len,hidden]
    # input_valid_mask [batch,len]  # 1 for nonpad 0 for pad

    with tf.name_scope(name):
        seq_len = tf.reduce_sum(input_valid_mask, axis=-1)  # [batch]

        # set pad vector to zero
        x = inputs * tf.expand_dims(input_valid_mask, -1)  # [batch,len,hidden]

        # mean pool
        x_mean = tf.reduce_sum(x, axis=1) / tf.cast(tf.expand_dims(seq_len, -1), tf.float32)

    return x_mean  # [batch, hidden]


def max_pool(inputs, input_valid_mask, name='max_pool'):
    # inputs [batch,len,hidden]
    # input_valid_mask [batch,len]  # 1 for nonpad 0 for pad

    with tf.name_scope(name):
        # set pad vector to negative infinity
        input_pad_mask = 1. - input_valid_mask
        x = inputs + tf.expand_dims(input_pad_mask * -1e9, -1)  # [batch,len,hidden]

        # max pool
        x_max = tf.reduce_max(x, axis=1)

    return x_max  # [batch, hidden]


def example_transformer_encoder():
    """ transformer encoder example code (just for reference) """
    encoder_input = None  # [batch,len,embed]
    dropout_rate = None
    encoder_valid_mask = mask_nonpad_from_embedding(encoder_input)  # [batch,len] 1 for nonpad; 0 for pad
    encoder_input = add_timing_signal_1d(encoder_input)  # add position embedding
    encoder_input = tf.layers.dropout(encoder_input, dropout_rate)
    encoder_output = transformer_encoder(encoder_input, encoder_valid_mask)


def example_transformer_decoder():
    """ transformer decoder example code (just for reference) """
    decoder_input = None  # [batch,len,embed]
    dropout_rate = None
    encoder_output = None  # [batch,len,hid]
    encoder_valid_mask = None  # [batch,len]

    decoder_valid_mask = mask_nonpad_from_embedding(decoder_input)  # [batch,len] 1 for nonpad; 0 for pad
    decoder_input = shift_right(decoder_input)  # 用pad当做eos
    decoder_input = add_timing_signal_1d(decoder_input)
    decoder_input = tf.layers.dropout(decoder_input, dropout_rate)
    decoder_output = transformer_decoder(decoder_input, encoder_output, decoder_valid_mask, encoder_valid_mask)


def example_transformer_encoder_decoder_train_and_infer():
    """ transformer encoder and decoder example code. especially decoder infer code (just for reference) """
    # encoder
    example_transformer_encoder()

    # decoder
    example_transformer_decoder()

    # transformer decoder infer
    # these args should stay the same with the decoder setting ahead
    encoder_output = None
    encoder_valid_mask = None
    num_decoder_layer = 6
    hidden_size = 512
    num_head = 6
    vocab_size = 10000
    embed_size = 512
    beam_size = 1  # greedy search
    beam_size = 10  # beam search
    max_decode_len = 50
    eos_id = 2

    batch_size = shape_list(encoder_output)[0]
    # 初始化缓存
    cache = {
        'layer_%d' % layer: {
            # 用以缓存decoder过程前面已计算的k,v
            'k': split_heads(tf.zeros([batch_size, 0, hidden_size]), num_head),
            'v': split_heads(tf.zeros([batch_size, 0, hidden_size]), num_head)
        } for layer in range(num_decoder_layer)
    }
    for layer in range(num_decoder_layer):
        # 对于decoder每层均需与encoder顶层隐状态计算attention,相应均有特定的k,v可缓存
        layer_name = 'layer_%d' % layer
        with tf.variable_scope('decoder/%s/encdec_attention/multihead_attention' % layer_name):
            k_encdec = tf.layers.dense(encoder_output, hidden_size, use_bias=False, name='k', reuse=tf.AUTO_REUSE)
            k_encdec = split_heads(k_encdec, num_head)
            v_encdec = tf.layers.dense(encoder_output, hidden_size, use_bias=False, name='v', reuse=tf.AUTO_REUSE)
            v_encdec = split_heads(v_encdec, num_head)
        cache[layer_name]['k_encdec'] = k_encdec
        cache[layer_name]['v_encdec'] = v_encdec
    cache['encoder_output'] = encoder_output
    cache['encoder_mask'] = encoder_valid_mask

    position_embedding = get_timing_signal_1d(max_decode_len, hidden_size)  # position embedding +eos [1,length+1,embed]

    def symbols_to_logits_fn(ids, i, cache):
        ids = ids[:, -1:]  # [batch,1] 截取最后一个
        target_embed, _ = embedding(tf.expand_dims(ids, axis=-1), vocab_size, embed_size, 'embedding')  # [batch,1,hidden] scope name需与前面保持一致

        decoder_input = target_embed + position_embedding[:, i:i + 1, :]  # [batch,1,hidden]

        encoder_output = cache['encoder_output']
        encoder_mask = cache['encoder_mask']

        with tf.variable_scope('', reuse=tf.AUTO_REUSE):
            decoder_output = transformer_decoder(decoder_input, encoder_output, None, encoder_mask,
                                                 cache=cache,  # 注意infer要cache
                                                 hidden_size=hidden_size,
                                                 filter_size=hidden_size * 4,
                                                 num_heads=num_head,
                                                 num_decoder_layers=num_decoder_layer,
                                                 )
        logits = proj_logits(decoder_output, embed_size, vocab_size, name='embedding')  # [batch,1,vocab]
        ret = tf.squeeze(logits, axis=1)  # [batch,vocab]
        return ret, cache

    initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

    def greedy_search_wrapper():
        """ Greedy Search """
        decoded_ids, scores = greedy_search(
            symbols_to_logits_fn,
            initial_ids,
            max_decode_len,
            cache=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    def beam_search_wrapper():
        """ Beam Search """
        decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
            symbols_to_logits_fn,
            initial_ids,
            beam_size,
            max_decode_len,
            vocab_size,
            alpha=0,
            states=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    decoded_ids, scores = tf.cond(tf.equal(beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

    # decoded_ids  # [batch,beam/1,len]
    # scores  # [batch,beam/1]


def rnn_decoder_train(decoder_input, encoder_output, decoder_valid_mask, encoder_valid_mask,
                      hidden_size, dropout_rate, num_decoder_layers=1,
                      cell_name='GRUCell', attention_name='BahdanauAttention'):
    """ use tf attention wrapper """
    # encoder_output: [batch,len,hid]
    # encoder_valid_mask: [batch,len]
    # cell_name: GRUCell or LSTMCell
    # attention_name: BahdanauAttention or LuongAttention
    encoder_output_seqlen = tf.cast(tf.reduce_sum(encoder_valid_mask, -1), tf.int32)  # [batch]

    Cell = getattr(tf.nn.rnn_cell, cell_name)  # class GRUCell/LSTMCell
    layers = [tf.nn.rnn_cell.DropoutWrapper(Cell(hidden_size), input_keep_prob=1.0 - dropout_rate) for _ in range(num_decoder_layers)]

    decoder_input = shift_right(decoder_input)  # 用0当做eos
    decoder_input_seqlen = tf.cast(tf.reduce_sum(decoder_valid_mask, -1), tf.int32)  # [batch]

    attention_mechanism = getattr(tf.contrib.seq2seq, attention_name)(
        hidden_size,
        encoder_output,
        memory_sequence_length=encoder_output_seqlen
    )

    cell = tf.contrib.seq2seq.AttentionWrapper(
        tf.nn.rnn_cell.MultiRNNCell(layers),
        attention_mechanism
    )

    with tf.variable_scope('decoder'):
        output, _ = tf.nn.dynamic_rnn(
            cell,
            decoder_input,
            decoder_input_seqlen,
            initial_state=None,  # 默认用0向量初始化
            dtype=tf.float32,
            time_major=False
        )  # 默认scope是rnn e.g.decoder/rnn/kernal

    return output


def rnn_decoder_infer(encoder_output, encoder_valid_mask,
                      hidden_size, dropout_rate, beam_size, vocab_size, num_decoder_layers=1,
                      cell_name='GRUCell', attention_name='BahdanauAttention', max_decode_len=50, eos_id=2):
    """ use tf attention wrapper infer """
    # encoder_output: [batch,len,hid]
    # encoder_valid_mask: [batch,len]
    # cell_name: GRUCell or LSTMCell
    # attention_name: BahdanauAttention or LuongAttention

    encoder_output_seqlen = tf.cast(tf.reduce_sum(encoder_valid_mask, -1), tf.int32)  # [batch]

    Cell = getattr(tf.nn.rnn_cell, cell_name)  # class GRUCell/LSTMCell
    layers = [tf.nn.rnn_cell.DropoutWrapper(Cell(hidden_size), input_keep_prob=1.0 - dropout_rate) for _ in range(num_decoder_layers)]

    batch_size = shape_list(encoder_output)[0]

    attention_mechanism = getattr(tf.contrib.seq2seq, attention_name)(
        hidden_size,
        encoder_output,
        memory_sequence_length=encoder_output_seqlen
    )

    cell = tf.contrib.seq2seq.AttentionWrapper(
        tf.nn.rnn_cell.MultiRNNCell(layers),
        attention_mechanism,
        output_attention=False
    )

    initial_state = cell.zero_state(batch_size * beam_size, tf.float32)  # 内部会检查batch_size与encoder_output是否一致,需乘beam_size

    # 区分能否设在cache: cache的值在beam_search过程中会expand和merge,需要tensor rank大于1
    cache = {
        'cell_state': initial_state.cell_state,
        'attention': initial_state.attention,
        'alignments': initial_state.alignments,
        'attention_state': initial_state.attention_state,
    }
    unable_cache = {
        'alignment_history': initial_state.alignment_history,
        'time': initial_state.time
    }

    # 将cache先变回batch,beam_search过程会expand/merge/gather,使得state是符合batch*beam的
    cache = nest.map_structure(lambda s: s[:batch_size], cache)

    def symbols_to_logits_fn(ids, i, cache):
        nonlocal unable_cache
        ids = ids[:, -1:]
        targets = tf.expand_dims(ids, axis=-1)  # [batch,1,1]
        embedding_target = embedding(targets, vocab_size, hidden_size, 'shared_embedding')

        decoder_input = tf.squeeze(embedding_target, axis=1)  # [batch,hidden]
        # 合并 cache和unable_cache为state
        state = cell.zero_state(batch_size * beam_size, tf.float32).clone(
            cell_state=cache['cell_state'],
            attention=cache['attention'],
            alignments=cache['alignments'],
            attention_state=cache['attention_state'],
            alignment_history=unable_cache['alignment_history'],
            time=unable_cache['time'],
        )

        with tf.variable_scope('decoder/rnn'):
            output, state = cell(decoder_input, state)
        # 分开cache和unable_cache
        cache['cell_state'] = state.cell_state
        cache['attention'] = state.attention
        cache['alignments'] = state.alignments
        cache['attention_state'] = state.attention_state
        unable_cache['alignment_history'] = state.alignment_history
        unable_cache['time'] = state.time
        body_output = output  # [batch,hidden]

        logits = proj_logits(body_output, hidden_size, vocab_size, name='shared_embedding')
        return logits, cache

    initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

    def greedy_search_wrapper():
        """ Greedy Search """
        decoded_ids, scores = greedy_search(
            symbols_to_logits_fn,
            initial_ids,
            max_decode_len,
            cache=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    def beam_search_wrapper():
        """ Beam Search """
        decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
            symbols_to_logits_fn,
            initial_ids,
            beam_size,
            max_decode_len,
            vocab_size,
            alpha=0,
            states=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    decoded_ids, scores = tf.cond(tf.equal(beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

    # decoded_ids  # [batch,beam/1,len]
    # scores  # [batch,beam/1]


def example_rnn_decoder_train_infer():
    encoder_output = None
    s1_seqlen = None
    s2 = None
    dropout_rate = 0.
    hidden_size, vocab_size, embed_size = None, None, None
    batch_size = shape_list(encoder_output)[0]
    beam_size = None,
    max_decode_len = 50,
    eos_id = 2
    # decoder
    decoder_rnn = getattr(tf.nn.rnn_cell, 'GRUCell')(hidden_size)  # GRUCell/LSTMCell
    encdec_atten = EncDecAttention(encoder_output, s1_seqlen, hidden_size)

    s2_embed, _ = embedding(tf.expand_dims(s2, -1), vocab_size, embed_size, name='share_embedding')
    s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
    s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]

    decoder_input = s2_embed
    decoder_input = shift_right(decoder_input)  # 用pad当做eos
    decoder_input = tf.layers.dropout(decoder_input, rate=dropout_rate)  # dropout

    init_decoder_state = decoder_rnn.zero_state(batch_size, tf.float32)
    # init_decoder_state = tf.nn.rnn_cell.LSTMStateTuple(c, h)

    time_step = tf.constant(0, dtype=tf.int32)
    rnn_output = tf.zeros([batch_size, 0, hidden_size])
    context_output = tf.zeros([batch_size, 0, hidden_size * 2])  # 注意力

    def loop_condition(time_step, *_):
        return tf.less(time_step, tf.reduce_max(s2_seqlen))

    def loop_body(time_step, prev_rnn_state, rnn_output, context_output):
        # attention
        s = prev_rnn_state if isinstance(decoder_rnn, tf.nn.rnn_cell.GRUCell) else prev_rnn_state.h
        context = encdec_atten(s)  # [batch,hidden]
        context_output = tf.concat([context_output, tf.expand_dims(context, axis=1)], axis=1)

        # construct rnn input
        rnn_input = tf.concat([decoder_input[:, time_step, :], context], axis=-1)  # [batch,hidden+]  use attention
        # rnn_input = decoder_input[:, time_step, :]  # [batch,hidden]  not use attention

        # run rnn
        current_output, rnn_state = decoder_rnn(rnn_input, prev_rnn_state)

        # append to output bucket via length dim
        rnn_output = tf.concat([rnn_output, tf.expand_dims(current_output, axis=1)], axis=1)

        return time_step + 1, rnn_state, rnn_output, context_output

    # start loop
    final_time_step, final_state, rnn_output, context_output = tf.while_loop(
        loop_condition,
        loop_body,
        loop_vars=[time_step, init_decoder_state, rnn_output, context_output],
        shape_invariants=[
            tf.TensorShape([]),
            nest.map_structure(get_state_shape_invariants, init_decoder_state),
            tf.TensorShape([None, None, hidden_size]),
            tf.TensorShape([None, None, hidden_size * 2]),
        ])
    # body_output = tf.concat([rnn_output, context_output], axis=-1)
    # body_output = tf.layers.dense(body_output, self.hidden_size, activation=tf.nn.tanh, use_bias=True, name='body_output_layer')
    decoder_output = rnn_output

    logits = proj_logits(decoder_output, embed_size, vocab_size, name='share_embedding')
    # logits = proj_logits(encoder_output[:,:,:300], embed_size, vocab_size, name='share_embedding')

    onehot_s2 = tf.one_hot(s2, depth=vocab_size)  # [batch,len,vocab]

    xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
    weights = tf.to_float(tf.not_equal(s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

    loss_num = xentropy * weights  # [batch,len]
    loss_den = weights  # [batch,len]

    loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar

    # decoder infer 
    cache = {'state': decoder_rnn.zero_state(batch_size, tf.float32)}

    def symbols_to_logits_fn(ids, i, cache):
        # ids [batch,length]
        pred_target = ids[:, -1:]  # 截取最后一个  [batch,1]
        embed_target, _ = embedding(tf.expand_dims(pred_target, axis=-1), vocab_size, embed_size, 'share_embedding')  # [batch,length,embed]
        decoder_input = tf.squeeze(embed_target, axis=1)  # [batch,embed]

        # if use attention
        s = cache['state'] if isinstance(decoder_rnn, tf.nn.rnn_cell.GRUCell) else cache['state'].h
        context = encdec_atten(s, beam_size=beam_size)  # [batch,hidden]
        decoder_input = tf.concat([decoder_input, context], axis=-1)  # [batch,hidden+]

        # run rnn
        # with tf.variable_scope('rnn', reuse=tf.AUTO_REUSE):
        decoder_output, cache['state'] = decoder_rnn(decoder_input, cache['state'])

        logits = proj_logits(decoder_output, hidden_size, vocab_size, name='share_embedding')

        return logits, cache

    initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

    def greedy_search_wrapper():
        """ Greedy Search """
        decoded_ids, scores = greedy_search(
            symbols_to_logits_fn,
            initial_ids,
            max_decode_len,
            cache=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    def beam_search_wrapper():
        """ Beam Search """
        decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
            symbols_to_logits_fn,
            initial_ids,
            beam_size,
            max_decode_len,
            vocab_size,
            alpha=0,
            states=cache,
            eos_id=eos_id,
        )
        return decoded_ids, scores

    decoded_ids, scores = tf.cond(tf.equal(beam_size, 1), greedy_search_wrapper, beam_search_wrapper)


class EncDecAttention():
    """ encoder decoder Bahdanauattention 可以将key进行缓存 """

    # memory -> key
    # query <-> key -> atten
    # att * memory -> ctx
    def __init__(self, memory, memory_length, attention_size):
        # memory [batch,len,memory_size(hidden)]
        # memory_length [batch] == seq_length
        with tf.variable_scope('encdec_attention', reuse=tf.AUTO_REUSE):
            self.attention_size = attention_size
            self.memory = memory
            self.memory_length = memory_length
            self.processed_key = tf.layers.dense(self.memory, self.attention_size, activation=None, use_bias=False, name='key_layer')  # [batch,len,atten]
            self.atten_v = tf.get_variable('attention_v', shape=[self.attention_size], dtype=self.memory.dtype)

    def __call__(self, query, beam_size=None):
        """ beam_size is not None 代表query的batch扩为batch*beam了,相应key也要扩"""
        if beam_size is not None:  # 支持beam_size = Tensor(1) Tensor不能直接==None
            processed_key_shape = shape_list(self.processed_key)
            processed_key = tf.expand_dims(self.processed_key, axis=1)  # [batch,1,len,hidden]
            processed_key = tf.tile(processed_key, [1, beam_size, 1, 1])  # [batch,beam,len,hidden]
            processed_key = tf.reshape(processed_key, [-1, processed_key_shape[1], processed_key_shape[2]])  # [batch*beam,len,hidden]

            memory_length = tf.expand_dims(self.memory_length, axis=1)  # [batch,1]
            memory_length = tf.tile(memory_length, [1, beam_size])  # [batch,beam]
            memory_length = tf.reshape(memory_length, [-1])  # [batch*beam]
        else:
            processed_key = self.processed_key
            memory_length = self.memory_length

        # query [batch,hidden]
        with tf.variable_scope('encdec_attention', reuse=tf.AUTO_REUSE):
            processed_query = tf.layers.dense(query, self.attention_size, activation=None, use_bias=False, name='query_layer')  # [batch,atten]
        processed_query = tf.expand_dims(processed_query, axis=1)  # [batch,1,atten]

        # attention alpha before normalizer
        score = tf.reduce_sum(self.atten_v * tf.tanh(processed_key + processed_query), axis=2)  # [batch,len] 加性注意力
        # tf.boolean_mask()
        # mask [batch,length]
        score_mask = tf.sequence_mask(memory_length, maxlen=tf.shape(score)[1])  # bool
        score_mask_values = -1e7 * tf.ones_like(score)
        score = tf.where(score_mask, score, score_mask_values)

        # normalizer
        alignments = tf.nn.softmax(score)  # [batch,len]

        # context c
        context = tf.expand_dims(alignments, axis=-1) * self.memory  # [batch,length,memory_size]
        context = tf.reduce_sum(context, axis=1)  # [batch,memory_size]
        return context  # [batch,memory_size]


def multihead_attention_encoder(query, key, key_valid_mask,
                                hidden_size, num_heads, dropout_rate,
                                name='multihead_attention', reuse=tf.AUTO_REUSE,
                                dropout_broadcast_dims=None):
    with tf.variable_scope(name, reuse=reuse):
        key_pad_mask = 1. - key_valid_mask  # 1 for pad 0 for nonpad 用于ffn  [batch,len]
        key_attention_bias = key_pad_mask * -1e9  # attention mask
        key_attention_bias = tf.expand_dims(tf.expand_dims(key_attention_bias, axis=1), axis=1)  # [batch,1,1,len]
        # key_attention_bias [batch,1,1,len_k] + [batch,head,len_q,len_k]
        output = multihead_attention(
            query,
            key,
            key_attention_bias,
            hidden_size,
            hidden_size,
            hidden_size,
            num_heads,
            dropout_rate,
        )
        output += query

        output = layer_norm(output)

        output = dropout_with_broadcast_dims(output, 1.0 - dropout_rate, broadcast_dims=dropout_broadcast_dims)

    return output


def conv_multi_kernel(inputs, filters, kernels_size=(1, 2, 3, 4), strides=1, padding='SAME',
                      name='conv_multi_kernel', reuse=tf.AUTO_REUSE,
                      activation=None, use_bias=False,
                      layer_norm_finally=True,
                      ):
    # mrfn function
    # inputs [batch,len,hid]
    with tf.variable_scope(name, reuse=reuse):
        output_lst = []
        for i, kernel_size in enumerate(kernels_size):
            output = tf.layers.conv1d(inputs, filters, kernel_size, strides=strides, padding=padding,
                                      activation=activation, use_bias=use_bias, name='conv1d_%d' % i)
            output_lst.append(output)
        output = tf.concat(output_lst, -1)  # [batch,len,filter*len(kernel)
        if layer_norm_finally:
            output = layer_norm(output)
        return output


def batch_coattention_nnsubmulti(utterance, response, utterance_mask, scope="co_attention", reuse=None):
    # mrfn function
    with tf.variable_scope(scope, reuse=reuse):
        dim = utterance.get_shape().as_list()[-1]
        weight = tf.get_variable('Weight', shape=[dim, dim], dtype=tf.float32)
        e_utterance = tf.einsum('aij,jk->aik', utterance, weight)
        a_matrix = tf.matmul(response, tf.transpose(e_utterance, perm=[0, 2, 1]))
        reponse_atten = tf.matmul(masked_softmax(a_matrix, utterance_mask), utterance)
        feature_mul = tf.multiply(reponse_atten, response)
        feature_sub = tf.subtract(reponse_atten, response)
        feature_last = tf.layers.dense(tf.concat([feature_mul, feature_sub], axis=-1), dim, use_bias=True, activation=tf.nn.relu, reuse=reuse)
    return feature_last


def masked_softmax(scores, mask):
    # mrfn function
    numerator = tf.exp(scores - tf.reduce_max(scores, 2, keepdims=True)) * tf.expand_dims(mask, axis=1)
    denominator = tf.reduce_sum(numerator, 2, keepdims=True)
    weights = tf.div(numerator + 1e-5 / tf.cast(tf.shape(mask)[-1], tf.float32), denominator + 1e-5)
    return weights



#!/usr/bin/env python
# coding=utf-8
import tensorflow as tf


class Bi_RNN():
    def __init__(self, **kwargs):
        """
        """
        self.cell_name = kwargs['cell_name']  # 'GRUCell'/'LSTMCell'
        self.dropout = kwargs['dropout_rate']  # default 0. (0,1)
        self.hidden_size = kwargs['hidden_size']

        Cell = getattr(tf.nn.rnn_cell, self.cell_name)  # class GRUCell/LSTMCell
        self.fw_cell = Cell(self.hidden_size, name='fw')
        self.bw_cell = Cell(self.hidden_size, name='bw')

        if isinstance(self.dropout, tf.Tensor):  # 是placeholder则一定要先wrapper了
            self.fw_cell = tf.nn.rnn_cell.DropoutWrapper(self.fw_cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)
            self.bw_cell = tf.nn.rnn_cell.DropoutWrapper(self.bw_cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)
        else:
            if self.dropout:  # float not 0.
                self.fw_cell = tf.nn.rnn_cell.DropoutWrapper(self.fw_cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)
                self.bw_cell = tf.nn.rnn_cell.DropoutWrapper(self.bw_cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)

    def __call__(self, embedding, seq_length, name=None, reuse=tf.AUTO_REUSE, **kwargs):
        """
        :param inputs embedding: [batch, length, embed] 
        seq_length: [batch]
        :return: outputs: [batch, length, 2*hidden], state: [batch, 2*hidden]
        """
        scope_name = name + '/' + f'bi_{self.cell_name}_encoder' if name else f'bi_{self.cell_name}_encoder'
        with tf.variable_scope(scope_name, reuse=reuse):
            (fw_outputs, bw_outputs), (fw_state, bw_state) = tf.nn.bidirectional_dynamic_rnn(self.fw_cell, self.bw_cell,
                                                                                             embedding,
                                                                                             sequence_length=seq_length,
                                                                                             dtype=tf.float32)
            if self.cell_name == 'LSTMCell':
                # LSTMStateTuple
                fw_state = fw_state.h
                bw_state = bw_state.h
            
            outputs = tf.concat([fw_outputs, bw_outputs], axis=-1)  # [batch,length,2*hidden]
            state = tf.concat([fw_state, bw_state], axis=-1)  # [batch,2*hidden]

            return outputs, state


class RNN():
    def __init__(self, **kwargs):
        """
        """
        self.cell_name = kwargs['cell_name']  # 'GRUCell'/'LSTMCell'
        self.dropout = kwargs['dropout_rate']  # default 0. (0,1)
        self.hidden_size = kwargs['hidden_size']
        self.name = kwargs['name']

        Cell = getattr(tf.nn.rnn_cell, self.cell_name)  # class GRUCell/LSTMCell
        self.cell = Cell(self.hidden_size, name='rnn_cell')

        if isinstance(self.dropout, tf.Tensor):  # 是placeholder则一定要先wrapper了
            self.cell = tf.nn.rnn_cell.DropoutWrapper(self.cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)
        else:
            if self.dropout:  # float not 0.
                self.cell = tf.nn.rnn_cell.DropoutWrapper(self.cell, input_keep_prob=1. - self.dropout, output_keep_prob=1.0)

    def __call__(self, embedding, seq_length, reuse=tf.AUTO_REUSE, one_step=False, **kwargs):
        """
        :param inputs embedding: [batch, length, embed] 
        :return: outputs: [batch, length, 2*hidden], state: [batch, 2*hidden]
        """
        with tf.variable_scope(f'{self.name}_{self.cell_name}', reuse=reuse):
            outputs, state = tf.nn.dynamic_rnn(self.cell,
                                               embedding,
                                               sequence_length=seq_length,
                                               dtype=tf.float32)
            return outputs, state  # [batch,length,hidden] [batch,hidden]

    def one_step(self, one_step_input, state):
        with tf.variable_scope(f'{self.name}_{self.cell_name}/rnn', reuse=tf.AUTO_REUSE):
            output, state = self.cell(one_step_input, state)
        return output, state  # [batch,hidden] [batch,hidden]



#!/usr/bin/env python
# coding=utf-8
import os
import tensorflow as tf
from .bert import modeling
from .bert import tokenization

class BERT(object):
    def __init__(self,
                 bert_model_dir,
                 is_training,
                 input_ids,  # [batch,len]
                 input_mask=None,  # [batch,len]
                 segment_ids=None,  # [batch,len]
                 verbose=False
                 ):
        self.bert_config = modeling.BertConfig.from_json_file(os.path.join(bert_model_dir, 'bert_config.json'))
        self.is_training = is_training
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.bm = None
        self.build_model(self.input_ids, self.input_mask, self.segment_ids)
        if self.is_training:
            self.restore_vars(os.path.join(bert_model_dir, 'bert_model.ckpt'), verbose=verbose)

    def build_model(self, input_ids, input_mask=None, segment_ids=None):
        self.bm = modeling.BertModel(
            config=self.bert_config,
            is_training=self.is_training,
            input_ids=input_ids,
            input_mask=input_mask,
            token_type_ids=segment_ids,
            use_one_hot_embeddings=False
        )

    def get_pooled_output(self):
        """ 获取对应的bert的[CLS]输出[batch_size, hidden_size] """
        if not self.bm:
            print('bert model has not build yet, please call build_model()')
        return self.bm.get_pooled_output()

    def get_sequence_output(self):
        """ 获取对应的bert输出[batch_size, seq_length, hidden_size] """
        if not self.bm:
            print('bert model has not build yet, please call build_model()')
        return self.bm.get_sequence_output()

    def get_seqlen(self):
        """ 获取input_ids真实长度 """
        seqlen = tf.cast(tf.reduce_sum(tf.sign(tf.abs(self.input_ids)), axis=1), tf.int32)  # [batch]
        return seqlen

    def get_nonpad_mask(self):
        """ 获取input_ids的mask: 1 for nonpad(!=0) - 0 for pad(=0) """
        mask = tf.sign(tf.abs(self.input_ids))  # [batch,len]
        return mask

    def restore_vars(self, init_checkpoint, verbose=False):
        """ 加载预训练的BERT模型参数 """
        print(f'<<<<<< restoring bert vars from ckpt: {init_checkpoint}')
        tvars = tf.trainable_variables()
        assignment_map, initialized_variable_names = modeling.get_assignment_map_from_checkpoint(tvars, init_checkpoint)
        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)
        if verbose:
            print("**** Trainable Variables ****")
            for var in tvars:
                init_string = ""
                if var.name in initialized_variable_names:
                    init_string = ", *INIT_FROM_CKPT*"
                print(f'  name = {var.name}, shape = {var.shape}{init_string}')

def get_tokenizer(vocab_file):
    return tokenization.FullTokenizer(vocab_file=vocab_file, do_lower_case=True)